import os
from math import pi
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.utils.data import Dataset, DataLoader

class Mish(nn.Module):
    def __init__(self):
        super(Mish, self).__init__()
    def forward(self, input):
        output = input*(torch.tanh(F.softplus(input)))
        return output

class Res_Block(nn.Module):
    def __init__(self,inc,outc,dosp = False):
        super(Res_Block, self).__init__()
        self.inc = inc
        self.outc = outc
        self.dosp = dosp
        self.layer0 = nn.Sequential(
            nn.Conv2d(in_channels=inc,out_channels=outc,kernel_size=3,
                   stride=1,padding=1,bias=False),
            nn.BatchNorm2d(outc),
            Mish()
        )
        if dosp:
            self.sample = nn.Upsample(scale_factor=0.5, mode='bilinear')
            self.layer1 = nn.Sequential(
                    nn.Conv2d(in_channels=outc, out_channels=outc, kernel_size=2,
                              stride=2, padding=0, bias=False),
                    nn.BatchNorm2d(outc))
        else:
            self.layer1 = nn.Sequential(
                    nn.Conv2d(in_channels=outc, out_channels=outc, kernel_size=3,
                              stride=1, padding=1, bias=False),
                    nn.BatchNorm2d(outc))
        self.mish = Mish()
        if inc!= outc:
            self.shortcut = nn.Conv2d(inc,outc,1,1,0)
    def forward(self, input):
        output0 = self.layer0(input)
        output0 = self.layer1(output0)
        if self.dosp:
            input = self.sample(input)
        if self.inc == self.outc:
            output1 = input
        else:
            output1 = self.shortcut(input)
        output = output0 + output1
        output = self.mish(output)
        return output




class Judge_Block(nn.Module):
    def __init__(self,outc):
        super(Judge_Block, self).__init__()
        self.layer0 = nn.Sequential(
                nn.Conv2d(512,512,1,bias=False),
                nn.BatchNorm2d(512),
                Mish(),
                nn.Conv2d(512,1024,2,bias=False),
                nn.BatchNorm2d(1024))
        self.sc = nn.Conv2d(512,1024,1)
        self.sample = nn.Upsample(scale_factor=0.5, mode='bilinear')
        self.mish = Mish()
        self.layer1 = nn.Conv2d(1024,outc,1)
    def forward(self, input):
        output0 = self.layer0(input)
        output1 = self.sample(input)
        output1 = self.sc(output1)
        output = output0 + output1
        output = self.mish(output)
        output = self.layer1(output)
        return output
    

    
class Model_2(nn.Module):
    def __init__(self):
        super(Model_2, self).__init__()
        self.layer0 = nn.Sequential(
                Res_Block(2,32,False),  # take origin far-field and modulated far-field as input
                Res_Block(32,64,True),
                Res_Block(64,64,False),
                Res_Block(64,128,True))
        self.layer1 = nn.Sequential(
                Res_Block(128,128,False),
                Res_Block(128,256,True),
                Res_Block(256,256,False),
                Res_Block(256,512,True))
        self.layer2 = Judge_Block(6)    # predict 6(7-1) relative phase
    def forward(self, input):
        output = self.layer0(input)
        output = self.layer1(output)
        output = self.layer2(output)
        return output


class TrainImg_Dataset(Dataset):
    def __init__(self, img0_path, img1_path, label_path):
        super(TrainImg_Dataset, self).__init__()
        self.img0_path = img0_path    # origin far-field path
        self.img1_path = img1_path    # modulated far-field path
        self.label_file = open(os.path.join(label_path, r'train_label.txt'), 'r').readlines() # label path
        self.img2data = trans.Compose([
            trans.ToTensor()
        ])

    def __len__(self):
        return len(self.label_file)

    def __getitem__(self, item):
        label = self.label_file[item].strip().split(' ')
        name = label[0]

        img0 = Img.open(os.path.join(self.img0_path, name))
        imgdata0 = self.img2data(img0)
        img1 = Img.open(os.path.join(self.img1_path, name))
        imgdata1 = self.img2data(img1)

        ori_phai_list = []
        for i in range(1, len(label)):
            ori_phai_list.append(float(label[i]))

        img_data = torch.cat([imgdata0, imgdata1], 0)
        phai_list = torch.Tensor(ori_phai_list)
        return img_data, phai_list


class Trainer():

    def __init__(self, load_parm_path, save_parm_path, label_path, train_path0, train_path1):
        self.net = Model_2().cuda()
        self.optim = optim.Adam(self.net.parameters())
        self.save_parm_path = save_parm_path
        self.label_path = label_path
        self.train_path0 = train_path0
        self.train_path1 = train_path1
        self.loss_fn = nn.MSELoss()
        if os.path.exists(load_parm_path):
            self.net.load_state_dict(torch.load(load_parm_path))
        self.net.train()

    def train(self):
        for epoch in range(200):
            train_dataset = TrainImg_Dataset(self.train_path0, self.train_path1, self.label_path)
            dataloader = DataLoader(train_dataset, batch_size=60, num_workers=4, shuffle=True)
            for i, (_img_data, _phai_list) in enumerate(dataloader):
                img_data = Variable(_img_data).cuda()
                phai_list = Variable(_phai_list).cuda()

                output = self.net(img_data)
                output = output.squeeze(3).squeeze(2)
                output = torch.tanh(output)
                output = output + torch.ones_like(output)
                phai_loss = self.loss_fn(output, phai_list)
                self.optim.zero_grad()
                phai_loss.backward()
                self.optim.step()
                print(epoch, '------', i, '------', phai_loss.item())

                if (i + 1) % 100 == 0:
                    torch.save(self.net.state_dict(),
                               os.path.join(self.save_parm_path,
                                            r'{}'.format(self.alpha),
                                            r'{0}_model2.parm'.format(epoch // 10)))
                    print('save successfully')


if __name__ == '__main__':
    trainer = Trainer(r'load_parm_path', r'save_parm_path',
                      r'label path', r'origin far-field path', r'modulated far-field path')
    trainer.train()